

/*
 *	watershed.c : Watershed segmentation.
 */


/*		CRYSTAL IMAGE 0.9.0 BETA
		This program is copyrighted material.
		(C) 2002-2010 by Mark A. Haidekker, all rights reserved.
		
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE:

		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
		APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
		HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
		OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
		THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
		PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
		IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
		ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

		IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
		WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES AND/OR CONVEYS
		THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES, INCLUDING ANY
		GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
		USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED TO LOSS OF
		DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD
		PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS),
		EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
		SUCH DAMAGES.

		For the full text of the license, please see the file COPYING
		in the project's main directory, or read it on the Web at
		http://www.gnu.org/copyleft/gpl.html

 */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <stdio.h>
#include <string.h>
#include <math.h>

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "application.h"
#include "backgnd.h"
#include "defs.h"
#include "dlgbox.h"
#include "filter.h"
#include "findclust.h"
#include "functions.h"
#include "graphdsp.h"
#include "gtkhelpers.h"
#include "imgbuf.h"
#include "interface.h"
#include "morphop.h"
#include "roi.h"
#include "secondwnd.h"
#include "tiffstuff.h"
#include "watershed.h"


image_type edmimg;			/* Bad design. With only few modifications, this can become local */


/********************************************
/
/
/	Threshold image based on the dialog box
/	(i.e. values in thresh_return )
/	The result goes into dest
/
/
**********************************************/


void edm_thresh (image_type* src, image_type* dest, float thresh, int slice)
{
rgbtriplet rgb;
rgbtriplet* sp;
short* p;
int x,y;

	p = dest->data; dest->minval.imin=0; dest->maxval.imax=1;
	
	if (src->imgtype==B_RGB)
	{
		sp = src->data;
		for (x=0; x<src->xmax; x++)
			for (y=0; y<src->ymax; y++)
			{
				rgb = sp[x + src->xmax*y];
				p[x + dest->xmax*y] = 0;
				if ((rgb.red >= thresh) || (rgb.green >= thresh) || (rgb.blue >= thresh))
					p[x + dest->xmax*y] = 1;
				else
					p[x + dest->xmax*y] = 0;
			}
	}
	else
	{
		for (x=0; x<src->xmax; x++)
			for (y=0; y<src->ymax; y++)
				p[x + dest->xmax*y] = (readbuf_flt(src,x,y,slice)>=thresh ? 1 : 0 );
	}
}


/***********************************************************************************
/
/	Euclidean Distance Map:
/	Create a EDM from an image. Background is whatever is below threshold,
/	anything above will be converted into an EDM. If do_thresh==1,
/	the threshold values from the threshold dialog box (dlgbox.c) will
/	be used, otherwise, the image will be assumed to be a 0/1 image
/
/	Since the EDM is a helper image, the main image remains untouched.
/	NOTE THAT THE TOP-LEVEL CALLER IS RESPONSIBLE FOR FREEING edmimg !
/
***********************************************************************************/


void create_edm_circ (image_type* img, float thresh, int slice)
{

int x,y,d,x1,y1;
short* p;
short dist;

int r,r1,r2,r3;


	/* Pass 1: create the EDM template, convert background to 0
	    b_short assumes that all images are less than 32000x32000 pixels in size */
	
	if (allocate_image (&edmimg, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	
	edm_thresh (img, &edmimg, thresh, slice);

	/* Pass 2 - for each pixel !=0 run in ever increasing rectangles until background found */

	p = edmimg.data;
	for (x1=0; x1<img->xmax; x1++)
	{
		dp (1, "%3.0f\b\b\b",100.0*x1/img->xmax);
		for (y1=0; y1<img->ymax; y1++)
			if (p[x1 + img->xmax*y1])	/* a non-zero pixel */
			{
				dist = 0; d=1;		/* start with a 3x3 box */
				while ((!dist) && (d<edmimg.xmax))
				{

					/* We use the Svenesian circle algorithm here */
					
					x = d; y=0; r=d*d;
					while (x>=0)
					{
						
						/* Check the current point in all 4 quadrants */
						
						if ( (x1+x < img->xmax) && (y1+y < img->ymax) && (p[x1+x + img->xmax*(y1+y)]==0)) dist=d;
						if ( (x1-x >=0        ) && (y1+y < img->ymax) && (p[x1-x + img->xmax*(y1+y)]==0)) dist=d;
						if ( (x1+x < img->xmax) && (y1-y >= 0       ) && (p[x1+x + img->xmax*(y1-y)]==0)) dist=d;
						if ( (x1-x >=0        ) && (y1-y >= 0       ) && (p[x1-x + img->xmax*(y1-y)]==0)) dist=d;
						
						/* Move to the next pixel of the circle */
						
						r1 = ARITHM(x-1,y); 	/* step to the left */
						r2 = ARITHM(x-1,y+1);	/* Diagonal step */
						r3 = ARITHM(x,y+1);		/* setp up */
						if ( (abs(r1-r)<abs(r2-r)) && (abs(r1-r)<abs(r3-r)) )
						{
							x--;
						}
						else if ( (abs(r2-r)<abs(r1-r)) && (abs(r2-r)<abs(r3-r)) )
						{
							x--; y++;
						}
						else
						{
							y++;
						}
						
					}
					if (dist==0) d++; /* One circle finished. If not yet found, increase radius */
				
				
				}
				p[x1+edmimg.xmax*y1]=dist;
				if (dist>edmimg.maxval.imax) edmimg.maxval.imax=dist;
			
			}
	}
	dp (1, "100 done\n");
}



/***********************************************************************************
/
/	Euclidean Distance Map:
/	Create a EDM from an image. Background is whatever is below threshold,
/	This is the alternative approach of iterative erosion
/
/***********************************************************************************/



void create_edm_erode (image_type* img, float thresh, int slice)
{
image_type mask;
int x1,y1;
short *p, *mp;
int iter, cnt;
int r,pixcnt;


	/* Pass 1: create the EDM template, convert background to 0
	    b_short assumes that all images are less than 32000x32000 pixels in size */
	
	if (allocate_image (&edmimg, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	if (allocate_image (&mask, B_SHORT, img->xmax, img->ymax, 1)==-1)
		fatal_memerr();
	
	edm_thresh (img, &mask, thresh, slice);
	dp (1, "Creating EDM, iterative erosion method\n");

	/* Pass 2 - Perform iterative erosion */

	p = edmimg.data;
	mp = mask.data;
	iter=0;							/* Erosion iteration counter */

	do
	{
		iter++; cnt=0; pixcnt=0;

		/* First pass, find all boundary points */

		for (y1=0; y1<img->ymax; y1++)
			for (x1=0; x1<img->xmax; x1++)
			{
				if (mp[x1 + img->xmax*y1])				/* a non-zero pixel */
				{
					pixcnt++;							/* Keep track of an all-zero mask, although this is redundant */
					r = skel_numneigh (&mask,x1,y1);	/* In morphop.c, determine number of nonzero neighbors */
					if (r<8)							/* means: at least one nonzero neighbor */
						p[x1 + img->xmax*y1] = iter;	/* Mark for erosion. Don't erode yet */

				}
			}

		/* Second pass, erode all boundary points */

		for (y1=0; y1<img->ymax; y1++)
			for (x1=0; x1<img->xmax; x1++)
			{
				if (p[x1 + img->xmax*y1]==iter)				/* a pixel marked for erosion in this iteration */
				{
					mp[x1 + img->xmax*y1]=0;				/* Erode this pixel */
					cnt++;									/* And count the number eroded */
				}
			}
		dp (1, "EDM: iteration %d, %d points eroded\n",iter,cnt);

	}
	while ((cnt>0) || (pixcnt>0));			/* End iteration when all points have been eroded */
	make_minmax (&edmimg);
}




/* Switch function for EDM method. This function also performs merging with original 
	grayscale image if requested and scales everything back into a SHORT array. */

void create_edm (image_type* img, float thresh, float imgweight, int slice, int method)
{
float *p; short *p1;
float a,min,max,buf, ofs,scale;
int x,y;

	if (method)
		create_edm_circ (img, thresh, slice);
	else
		create_edm_erode (img, thresh, slice);

	/* At this point, the EDM is in edmimg. We are done if imgweight==0 */

	if (F_ISZERO(imgweight)) return;

	/* Here, we need to muddle edmimg and img. Important that img contains no negative values,
		so we offset img if it cotains negative values */

	p = calloc (img->xmax*img->ymax, sizeof (float));
	if (!p) fatal_memerr();
	a = imgweight/100.0;
	imgminmax (img, &min,&max); if (min<0) ofs=min; else ofs=0.0;
	if (max==min) scale=10.0; else scale = 10.0/(max-min);

	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			buf = (1.0-a)*10.0*readbuf_flt(&edmimg,x,y,0) + scale*a*readbuf_flt(&edmimg,x,y,0)*(readbuf_flt(img,x,y,slice)-ofs);
			if ((x==0) && (y==0)) { min=max=buf; }
			min = MIN(min,buf);
			max=MAX(max,buf);
			p[x+img->xmax*y] = buf;
		}

	/* Store back in edmimg */

	p1 = edmimg.data;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{
			p1[x+img->xmax*y] = (int)(0.5+p[x+img->xmax*y]);
		}
	make_minmax (&edmimg);


}


/* General driver for EDM that takes care of returning the EDM properly.
	It also emulates the Gaussian filtering of the UEP and Watershed functions.
	Elemsize is just a dummy; keeps the parameter list compatible */

void	make_edm (image_type* img, float thresh, int gauss_size, int elemsize, float imgweight, int slice, int method)
{

	create_edm (img ,thresh, imgweight, slice, method);
	if (gauss_size>0) gaussian_general (&edmimg, gauss_size, 0);
	handback (img, &edmimg);

}




/****************************************************************************
/
/	Watershed: Based on the EDM, we get the watershed through the following steps:
/
/	1) Find local maxima of the EDM (any pt whose neighbors have l or eq value)
/	2) Make those points negative
/	3) Dilate those points (swap sign of neighbors) if non-zero AND not touching 
/      another negative pixel
/	4) Remove all positive pixels, swap sign of all negative ones
/
************************************************************************************/


/* Check if point at (x,y) is local maximum, return 1 if yes.
   Env determines what "local" means.
	This function is similar, but not identical to, the 
	is_locmax function in filter.c: the definition here (maxval>curval)
	allows for plateaus of equal value */

char wshd_is_locmax (image_type* img, int x, int y, int env)
{
char ismax; float maxval, curval;
long idx; int x1,y1,x2,y2,i,j;
float buf;

	idx = x + img->xmax*y;
	curval = readbuf_flt(img,x,y,0);		/* Value at central pixel */
	if (F_ISZERO(curval)) return 0;			/* Background is never a local maximum */
	
	/* Clamp environment */
	x1=x-env; if (x1<0) x1=0;
	x2=x+env; if (x2>=img->xmax) x2=img->xmax-1;
	y1=y-env; if (y1<0) y1=0;
	y2=y+env; if (y2>=img->ymax) y2=img->ymax-1;

	
	/* Now get the maximum value in the env-sized neighborhood */
	ismax = 1; maxval = readbuf_flt (img,x1,y1,0);
	for (i=y1; i<=y2; i++)
		for (j=x1; j<=x2; j++)
		{
			buf=readbuf_flt(img,j,i,0);
			if (((i!=y) || (j!=x)) && (maxval< buf)) maxval=buf;
		}

	if (maxval>curval) ismax=0; 	/* use > to allow ridges and plateaus */

	return ismax;
}




/*	The following code finds the Ultimate Eroded Points, or UEPs. Since UEPs may
	form ridges in oval shapes, we need to combine multiple points of the same
	Euclidean distance in a cluster-like fashion. This will be performed by
	transferring the UEPs into an auxiliary array and flood-filling them.
	find_UEP is a 2D function and operates on Slice 0 only!
*/


void UEP_floodfill (image_type* src, short* dest, int x, int y, short val, short env)
{

	/* Abort recursion if out of range */
	if ((x<0) || (y<0) || (x>=src->xmax) || (y>=src->ymax)) return;
	
	/* Set the current point if it is a local maximum and has not yet been visited */
	
	if ((dest[x + src->xmax*y]==0) && (wshd_is_locmax (src,x,y,env)))
	{
		dest[x + src->xmax*y] = -val;
		
		/* and recurse to its eight neighbors */
		
		UEP_floodfill (src, dest, x-1,y, val, env);
		UEP_floodfill (src, dest, x+1,y, val, env);
		UEP_floodfill (src, dest, x,y-1, val, env);
		UEP_floodfill (src, dest, x,y+1, val, env);
		UEP_floodfill (src, dest, x-1,y-1, val, env);
		UEP_floodfill (src, dest, x+1,y-1, val, env);
		UEP_floodfill (src, dest, x-1,y+1, val, env);
		UEP_floodfill (src, dest, x+1,y+1, val, env);
	}

}

int find_UEP (image_type* img, int env)		/* Find ultimate eroded points = local maxima */	
{
int x,y;
short* aux;
short* p;
short val;
long idx;

	g_assert (img->imgtype==B_SHORT);
	p = img->data;
	aux = calloc (img->xmax*img->ymax, sizeof (short));
	if (!aux) fatal_memerr();
	
	val=0;
	for (y=1; y<img->ymax-1; y++)
	{
		for (x=1; x<img->xmax-1; x++)
		{
			if ((p[x+img->xmax*y]>0) && (aux[x + img->xmax*y]==0) && wshd_is_locmax (img,x,y,env))
			{
				val++; 
				UEP_floodfill (img, aux, x,y, val, env);
			}
		}
	}
	
	/* Copy the negative values of aux into the img buffer, overriding the local maxima */
	
	idx=0;
	for (y=1; y<img->ymax-1; y++)
	{
		idx = y*img->xmax;
		for (x=1; x<img->xmax-1; x++)
		{
			idx++; /* because we start with x=1 */
			if (aux[idx]<0) p[idx]=aux[idx];
		}
	}

	free (aux);
	return val+1;
}


/* Watershed-flip: The 3x3 vicinity of a flipped pixel (at x,y) is
	analyzed. Every pixel in this vicinity that
	- is not background (!=0 in the original bitmap)
	- has not been processed (==0 in the p1 bitmap)
	- does not have another flipped pixel in its vicinity
	will be flipped
*/

int ws_flip (image_type* img, short* p1, int iter, int xc, int yc, int foureight)
{
int x1,x2,y1,y2,x,y,x3,x4,y3,y4,X,Y,k;
short* p; short source, test;
long idx;
int flipped;
char may_flip;
char fofs[9];

	p = img->data; idx=xc + img->xmax * yc; 
	source = p[idx]; flipped = 0;
	
	/* Range-checked 3x3 environment. Either 4- or 8-neighborhood */
	
	for (k=0; k<9; k++) fofs[k]=1;
	k=0;
	for (y=yc-1; y<=yc+1; y++)
		for (x=xc-1; x<=xc+1; x++)
		{
			if ((x<0) || (y<0) || (x>=img->xmax)|| (y>=img->ymax)) fofs[k]=0;
			k++;
		}

	if (foureight==4)	/* Flag the corner elements */
	{
		fofs[0]=0; fofs[2]=0; fofs[6]=0; fofs[8]=0; 
	}
	
	
	x1=xc-1; if (x1<0) x1=0;
	x2=xc+1; if (x2>=img->xmax) x2=img->xmax-1;
	y1=yc-1; if (y1<0) y1=0;
	y2=yc+1; if (y2>=img->ymax) y2=img->ymax-1;

	/* Scan the 3x3 vicinity of the source pixel at xc,yc */

	k=0;
	for (y=y1; y<=y2; y++)
		for (x=x1; x<=x2; x++)
		{
			if ((fofs[k])					/* Not out of range */
				&& (p[x+img->xmax*y]!=0)		/* not a background pixel */
				&& (p1[x+img->xmax*y]==0)	/* not yet visited */
				&& (p[x+img->xmax*y]>=iter)) /* it's this pixel's turn */
			{
				/* test 3: No flipped value != source may be in its vicinity
				   -- always 8-neighborhood */
				
				x3=x-1; if (x3<0) x3=0;
				x4=x+1; if (x4>=img->xmax) x4=img->xmax-1;
				y3=y-1; if (y3<0) y3=0;
				y4=y+1; if (y4>=img->ymax) y4=img->ymax-1;
				may_flip=1;
				for (Y=y3; Y<=y4; Y++)
					for (X=x3; X<=x4; X++)
					{
						test=p1[X+img->xmax*Y];
						if ((test<0) && (test!=source)) may_flip=0;
					}
				
				/* Now dilate if we still may */
				
				if (may_flip) { p1[x+img->xmax*y] = source; flipped++; }
				
			
			}
			k++;
		}

	return flipped;
}





int watershed (image_type* img, float thresh, int gauss_size, int elemsize, float imgweight, int slice, int edm_mode)
{
int changes,cnt,foureight;
int iter,x,y;
short* p;
short* p1;
long idx;

	/* Step 1: Create an EDM from image, then perform a thorough
	   low-pass filtering step to grind down small local pseudo-maxima.
		Then search for local maxima which are the UEPs */

	create_edm (img, thresh, imgweight, slice, edm_mode);
	dp (1, "Watershed: EDM prepared\nFiltering EDM...");
	if (gauss_size>0) gaussian_general (&edmimg, gauss_size, 0);
	dp (1, "done\n");
	make_minmax (&edmimg);
	cnt=find_UEP (&edmimg, elemsize);						/* Flip sign of local maxima */
	p = edmimg.data;

	dp (1, "Watershed: %d local maxima found\n",cnt);
	p1 = (short*) calloc (edmimg.xmax*edmimg.ymax, sizeof (short));
	if (!p1) fatal_memerr();


	/* Step 2: Successive dilation: Iteratively flip any sign of a pixel
	   that is in the neighborhood of a pixel with already flipped 
	   sign */
	
	iter=edmimg.maxval.imax;
	do
	{
		changes=0; 
		
		/* Start with blank aux image, merely copying the already-flipped areas.
		   For the first iteration, these are just the UEPs. */
		
		for (idx=0; idx<edmimg.xmax*edmimg.ymax; idx++)
		{
			if (p[idx]<0) p1[idx]=p[idx]; else p1[idx]=0;
		}

		/* Run over the image, dilating */

		foureight=8;
		for (y=1; y<edmimg.ymax-1; y++)
		{
			idx = y*edmimg.xmax;
			for (x=1; x<edmimg.xmax-1; x++)
			{
				idx++;	/* because we start at x=1 */
				
				/* If we find a flipped pixel, flip its 3x3 environment in aux */
				if (p[idx]<0) changes+=ws_flip(&edmimg, p1, iter, x,y, foureight);
				foureight = 12-foureight;

			}
		}
		dp (1, "Watershed: Iteration %d, changes: %d\n",iter,changes);
		
		/* The newly found flipped pixels need to be copied into edmimg. */

		for (idx=0; idx<edmimg.xmax*edmimg.ymax; idx++)
		{
			if (p1[idx]<0) p[idx]=p1[idx];
		}
		
		if (iter>0) iter--;
	}
	while ((iter>0) || ((iter==0) && (changes)));

	free (p1);

	/* Post-processing: Save only pixels with negative sign.
	   The segmentation result is now in edmimg. */
	
	for (idx=0; idx<edmimg.xmax*edmimg.ymax; idx++)
	{
		if (p[idx]<0) 
		{
			p[idx] = -p[idx];
		}
		else 
			p[idx]=0;
	}
	edmimg.maxval.imax=cnt-1;
	handback (img, &edmimg);
	return cnt;
	
}


/* This is a verbatim copy of the watershed part up to the
   UEP generation. We need to filter out the negative elements
   to get an UEP map. */

int create_UEP (image_type* img, float thresh, int gauss_size, int elemsize, float imgweight, int slice, int edm_mode)
{
int cnt;
short* p;
short* p1; 
long idx;



	/* Step 1: Create an EDM from image, then perform a thorough
	   low-pass filtering step to grind down small local pseudo-maxima.
		Then search for local maxima which are the UEPs */

	create_edm (img, thresh, imgweight, slice, edm_mode);
	dp (1, "Watershed: EDM prepared\nFiltering EDM...");
	if (gauss_size>0) gaussian_general (&edmimg, gauss_size, 0);
	dp (1, "done\n");
	make_minmax (&edmimg);
	cnt=find_UEP (&edmimg, elemsize);						/* Flip sign of local maxima */
	p = edmimg.data;

	dp (1, "UEP: %d local maxima found\n",cnt);
	p1 = (short*) calloc (edmimg.xmax*edmimg.ymax, sizeof (short));
	if (!p1) fatal_memerr();

	for (idx=0; idx	< edmimg.xmax*edmimg.ymax; idx++)
	{
		if (p[idx]<0)
			p[idx]=-p[idx];
		else
			p[idx]=0;
	}
	
	free (p1);
	handback (img, &edmimg);

	return cnt;
	
}




/* A variation of the theme (used elsewhere):
	Find local maxima, but allow detection of plateaus.

	Since we *absolutely* require a single point per local maximum,
	we need to reduce each cluster down to ...something. In this case,
	we flood-fill the cluster and use its centroid as locmax.
	Therefore, we need a floodfill helper function similar to
	UEP_floodfill above. Since we are at it, we also delete all features
	that touch the edge of the image.

	This is a 2D function and operates on one slice only!
*/


void cog_floodfill (image_type* img, int x, int y, int *cx, int *cy, int *cnt, unsigned char *edge_flag)
{
unsigned char *dest = img->data;
	
	/* Abort recursion if out of range */
	if ((x<0) || (y<0) || (x>=img->xmax) || (y>=img->ymax)) 
	{
		*edge_flag=1;
		return;
	}
	
	/* Label the current pixel and recurse to its neighbors */
	
	if (dest[x + img->xmax*y]==255)
	{
		dest[x + img->xmax*y] = 127;
		*cx += x; *cy += y; *cnt+=1;

		/* and recurse to its eight neighbors */
		
		cog_floodfill (img, x-1,y, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x+1,y, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x,y-1, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x,y+1, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x-1,y-1, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x+1,y-1, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x-1,y+1, cx, cy, cnt, edge_flag);
		cog_floodfill (img, x+1,y+1, cx, cy, cnt, edge_flag);
	}

}



int locmax_uep (image_type* img, float thresh, int elemsize, int slice)
{
int cnt,x,y;
unsigned char edge_flag;
int centx, centy, area;
unsigned char *p;
long idx;
image_type uepimg;



	if (allocate_image (&uepimg, B_CHAR, img->xmax, img->ymax, 1) ==-1)
		fatal_memerr();
	p = uepimg.data;

	/* In analogy to the locmax filter in filter.c, find local maxima.
		Note that local maxima can be clusters (ridges, blobs etc) */

	cnt=0;
	for (y=0; y<img->ymax; y++)
		for (x=0; x<img->xmax; x++)
		{	/* Observe both the threshold criterion and the local maximum criterion */
			if ((readbuf_flt (img,x,y,slice) >= thresh) && wshd_is_locmax (img,x,y,elemsize))
			{
				dp (2, "Local maximum at %d,%d\n",x,y);
				p[x + y*uepimg.xmax] = 255;					/* Label that point */
				cnt++;
			}
		}
	dp (1, "%d raw local maxima found.\n",cnt);


	/* We now find seed points (value=255) and flood-fill the regions.
		Label cluster centroids with the value 1 */

	cnt=0;
	for (y=0; y<uepimg.ymax; y++)
		for (x=0; x<uepimg.xmax; x++)
		{
			if ( ((unsigned char*)uepimg.data)[x + uepimg.xmax * y] == 255)
			{
				centx=0; centy=0; area=0; edge_flag=0;
				cog_floodfill (&uepimg, x,y, &centx, &centy, &area, &edge_flag);
				dp (2, "Reduced local maximum at %d, %d (originally %d pixels)\n",centx, centy, area);
				if (area > 0)						/* Kludge! We should never get area==0 */
				{
					centx /= area; centy /= area; cnt++;
					((unsigned char*)uepimg.data)[centx + uepimg.xmax * centy] = 1+edge_flag;
				}
			}
		}
	dp (1, "%d total local maxima found.\n",cnt);


	/* Now all that's left is to delete everything in upeimg that is not 1 */

	for (y=0; y<uepimg.ymax; y++)
		for (x=0; x<uepimg.xmax; x++)
		{
			if ( ((unsigned char*)uepimg.data)[x + uepimg.xmax * y] > 1)
				((unsigned char*)uepimg.data)[x + uepimg.xmax * y] = 0;
		}
	uepimg.minval.imin = 0;
	uepimg.maxval.imax = 1;

	handback (img, &uepimg);

	return cnt;
	
}






/************************************************
/
/	GUI drivers for EDM and Watershed 
/ 	both based on thshold dialog
/
************************************************/


int wshd_edm_mode, wshd_function, wshd_elemsize, wshd_gauss_size;
float wshd_threshold, wshd_imgweight;





/* Driver for the Watershed functions, for mainimage only */

void wshddriver (float threshold, int function, int edm_mode)
{
	undo_nexus (&mainimg);

	if (function==0)									/* Watershed */
		watershed (&mainimg, threshold, wshd_gauss_size, wshd_elemsize, wshd_imgweight, cur_slice, edm_mode);
	else if (function==1)								/* EDM */
		make_edm (&mainimg, threshold, wshd_gauss_size, wshd_elemsize, wshd_imgweight, cur_slice, edm_mode);
	else if (function==2)
		create_UEP (&mainimg, threshold, wshd_gauss_size, wshd_elemsize, wshd_imgweight, cur_slice, edm_mode);
	else
		return;

	cur_slice=0;
	make_imlib_buffer_from_main_image (cur_slice);
	redraw (0);


}



void wshd_adjustment_changed (GtkWidget* widget, gpointer user_data)
{ 
int which_adj;
float buf;
GtkAdjustment *w;

	
	if (user_data) which_adj = *(int*)user_data; else which_adj=0;
	w = GTK_ADJUSTMENT(widget);			/* Funny way to typecast */

	buf= w->value;
	
	if (which_adj==1)
	{
		wshd_threshold = buf;
		dp (2, "Watershed: Threshold selected is %f\n",wshd_threshold);
	}
	else if (which_adj==2)
	{
		wshd_elemsize = buf;
		dp (2, "Watershed: Element size is now %d (%d by %d)\n",wshd_elemsize, 2*wshd_elemsize+1,2*wshd_elemsize+1);
	}
	else if (which_adj==3)
	{
		wshd_gauss_size = buf;
		if (wshd_gauss_size==0)
			dp (2, "Watershed: Gaussian filtering is now deactivated\n");
		else
			dp (2, "Watershed: Gaussian kernel size is now %d (%d by %d)\n",wshd_gauss_size, 2*wshd_gauss_size+1,2*wshd_gauss_size+1);
	}
	else if (which_adj==4)
	{
		wshd_imgweight = buf;
		dp (2, "Watershed: Weighting original image with %3.0f percent\n",wshd_imgweight);
	}


}




/* Note that the next function is used for both option menus */

void on_wshd_menu_changed (GtkWidget *widget, gpointer user_data)
{
int i;

	i = *(int*) user_data;
	if (i<10)
		wshd_function = i;
	else
		wshd_edm_mode = i-10;

}

void on_wshd_cancel_clicked  (GtkButton *button, gpointer user_data)
{

	gtk_widget_destroy (GTK_WIDGET(user_data));
}


void on_wshd_ok_clicked  (GtkButton *button, gpointer user_data)
{

	wshddriver (wshd_threshold, wshd_function, wshd_edm_mode);

	gtk_widget_destroy (GTK_WIDGET(user_data));

}


/* The general control GUI for Watershed (function==0), EDM (function==1), and UEP (function==2) */


void create_wshd_wnd (image_type* img)
{
  GtkWidget *wshd_wnd;
  GtkWidget *wshd_table1;
  GtkWidget *wshd_label1, *wshd_label2, *wshd_label3, *wshd_label4, *wshd_label5;
  GtkWidget *wshd_okbutton;
  GtkWidget *wshd_cancelbutton;
  GtkWidget *wshd_t;
  GtkAdjustment *wshd_t_adj;
  GtkWidget *hsep;
  GtkWidget *l2_menu, *l2_menu_menu;
  GtkWidget *l3_menu, *l3_menu_menu;
  GtkWidget *menuitem;
  GtkWidget *wshd_size;
  GtkAdjustment *wshd_size_adj;
  GtkWidget *wshd_gauss;
  GtkAdjustment *wshd_gauss_adj;
  GtkWidget *wshd_weight;
  GtkAdjustment *wshd_weight_adj;

  
  float min,max;
  int y;

	wshd_function=0;
	wshd_edm_mode=1;
	wshd_elemsize=8;
	wshd_gauss_size=0;
	wshd_imgweight=0;
	if (img->imgtype == B_FLOAT) max = img->maxval.fmax; else max = img->maxval.imax;
	if (img->imgtype == B_FLOAT) min = img->minval.fmin; else min = img->minval.imin;
	wshd_threshold = min+1;

	wshd_wnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_object_set_data (GTK_OBJECT (wshd_wnd), "wshdes_wnd", wshd_wnd);
	gtk_window_set_title (GTK_WINDOW (wshd_wnd), "Watershed Functions");
	gtk_container_set_border_width (GTK_CONTAINER (wshd_wnd), 3);
	gtk_window_set_modal (GTK_WINDOW (wshd_wnd), TRUE);

	wshd_table1 = gtk_table_new (8, 2, FALSE);
	gtk_widget_ref (wshd_table1);
	gtk_object_set_data_full (GTK_OBJECT (wshd_wnd), "wshd_table1", wshd_table1,
                              (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (wshd_table1);
	gtk_container_add (GTK_CONTAINER (wshd_wnd), wshd_table1);

	y=0;
	wshd_label1 = MakeLabel (wshd_wnd,wshd_table1,"Operation", "label1",0,y);
	l2_menu = Make_Optionmenu (wshd_wnd,wshd_table1,"l2_menu",1,y);
	l2_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_wshd_menu_changed), "Watershed", (gpointer) &menuopt0);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_wshd_menu_changed), "Euclidean Distance Map (EDM)", (gpointer) &menuopt1);
	menuitem = Option_menuitem (l2_menu_menu, GTK_SIGNAL_FUNC(on_wshd_menu_changed), "Ultimate Eroded Points (UEP)", (gpointer) &menuopt2);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (l2_menu), l2_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (l2_menu), wshd_function);

	y++;
	wshd_label2 = MakeLabel (wshd_wnd,wshd_table1,"EDM method", "label2",0,y);
	l3_menu = Make_Optionmenu (wshd_wnd,wshd_table1,"l3_menu",1,y);
	l3_menu_menu = gtk_menu_new ();
	menuitem = Option_menuitem (l3_menu_menu, GTK_SIGNAL_FUNC(on_wshd_menu_changed), "Iterative Erosion", (gpointer) &menuopt10);
	menuitem = Option_menuitem (l3_menu_menu, GTK_SIGNAL_FUNC(on_wshd_menu_changed), "Euclidean (slow)", (gpointer) &menuopt11);
	gtk_option_menu_set_menu (GTK_OPTION_MENU (l3_menu), l3_menu_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (l3_menu), wshd_edm_mode);

	y++;
	wshd_label2 = MakeLabel (wshd_wnd, wshd_table1, "Threshold", "wshd_label2", 0,y);
	wshd_t_adj = (GtkAdjustment*)gtk_adjustment_new (wshd_threshold, min, max, 1, 10, 0);
	wshd_t = MakeSpinbutton (GTK_ADJUSTMENT(wshd_t_adj), wshd_wnd, wshd_table1, "wshd_t", 1,y);
	gtk_signal_connect (GTK_OBJECT (wshd_t_adj), "value_changed", GTK_SIGNAL_FUNC (wshd_adjustment_changed), &menuopt1);

	y++;
	wshd_label3 = MakeLabel (wshd_wnd, wshd_table1, "Element size/2", "wshd_label3", 0,y);
	wshd_size_adj = (GtkAdjustment*)gtk_adjustment_new (wshd_elemsize, 1, 15, 1, 1, 0);
	wshd_size = MakeSpinbutton (GTK_ADJUSTMENT(wshd_size_adj), wshd_wnd, wshd_table1, "wshd_dummy", 1,y);
	gtk_signal_connect (GTK_OBJECT (wshd_size_adj), "value_changed", GTK_SIGNAL_FUNC (wshd_adjustment_changed), &menuopt2);

	y++;
	wshd_label4 = MakeLabel (wshd_wnd, wshd_table1, "Gaussian smoothing", "wshd_label4", 0,y);
	wshd_gauss_adj = (GtkAdjustment*)gtk_adjustment_new (wshd_gauss_size, 0, 15, 1, 1, 0);
	wshd_gauss = MakeSpinbutton (GTK_ADJUSTMENT(wshd_gauss_adj), wshd_wnd, wshd_table1, "wshd_gauss", 1,y);
	gtk_signal_connect (GTK_OBJECT (wshd_gauss_adj), "value_changed", GTK_SIGNAL_FUNC (wshd_adjustment_changed), &menuopt3);

	y++;
	wshd_label5 = MakeLabel (wshd_wnd, wshd_table1, "EDM <-> Gray info", "wshd_label5", 0,y);
	wshd_weight_adj = (GtkAdjustment*)gtk_adjustment_new (wshd_imgweight, 0, 100, 1, 10, 0);
	wshd_weight = MakeSpinbutton (GTK_ADJUSTMENT(wshd_weight_adj), wshd_wnd, wshd_table1, "wshd_weight", 1,y);
	gtk_signal_connect (GTK_OBJECT (wshd_weight_adj), "value_changed", GTK_SIGNAL_FUNC (wshd_adjustment_changed), &menuopt4);

	y++;
	hsep = MakeHsep (wshd_wnd, wshd_table1, "hseparator4", 0,y);
	hsep = MakeHsep (wshd_wnd, wshd_table1, "hseparator5", 1,y);

	y++;
	wshd_okbutton = Button_in_table (wshd_wnd, wshd_table1, "OK", "wshd_okbutton", 0,y, NULL);
	wshd_cancelbutton = Button_in_table (wshd_wnd, wshd_table1, "Cancel", "wshd_canbutton", 1,y, NULL);

	gtk_signal_connect (GTK_OBJECT (wshd_okbutton), "clicked", 
                    	GTK_SIGNAL_FUNC (on_wshd_ok_clicked), wshd_wnd);
	gtk_signal_connect (GTK_OBJECT (wshd_cancelbutton), "clicked", 
                    	GTK_SIGNAL_FUNC (on_wshd_cancel_clicked), wshd_wnd);
	gtk_signal_connect (GTK_OBJECT (wshd_wnd), "destroy",
                    	GTK_SIGNAL_FUNC (on_wshd_cancel_clicked), wshd_wnd);


	gtk_widget_show (wshd_wnd);
}





void watershed_main ()
{
	create_wshd_wnd (&mainimg);
}


/**********************************************************/


